//
//  rtc.c
//
//  Interrupt base Real time clock
//  Copyright(C)2013-2016 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2013.05.22 v0.0  First cording
//    2014.03.05 v0.1  gettimeofday()
//    2015.02.03 v0.2  Supported the PIC32MX, SYS_TMR version(bad precision)
//    2016.02.05 v0.3  Supported the PIC32MX, PLIB_TMR version
//    2016.04.03 v0.4  Supported the PIC32MX, DRV_TMR version(XC-32(gcc) option:define RTC_DRV_TMR=DRV_TMR_INDEX_?)
//
//
#include <time.h>
#include <sys/time.h>
#include <stddef.h>
#include <errno.h>
#ifndef EFAULT
#define EFAULT          14              /* Bad address */
#endif


//
//    Global area
//
volatile time_t        rtc_time;

//
//    Unix system call
//
time_t time(time_t *t) {
    if (t != NULL) {
        *t = rtc_time;
    }
    return(rtc_time);
}

//
//  STM32 Implementation
//
#ifdef USE_STDPERIPH_DRIVER
void rtc_open(void) {
	NVIC_InitTypeDef		n;
	TIM_TimeBaseInitTypeDef	t;

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM4, ENABLE);

	n.NVIC_IRQChannel                   = TIM4_IRQn;
	n.NVIC_IRQChannelPreemptionPriority = 0;
	n.NVIC_IRQChannelSubPriority        = 1;
	n.NVIC_IRQChannelCmd                = ENABLE;
	NVIC_Init(&n);

	TIM_InternalClockConfig(TIM4);
	t.TIM_Period        = 10000;	// 100us * 10000 = 1s
	t.TIM_Prescaler     = ((SystemCoreClock / 2) / 10000) - 1;	// 10kHz:100us;
	t.TIM_ClockDivision = 0;
	t.TIM_CounterMode   = TIM_CounterMode_Up;
	TIM_TimeBaseInit(TIM4, &t);
	TIM_ITConfig(TIM4, TIM_IT_Update, ENABLE);
	TIM_Cmd(TIM4, ENABLE);
	return;
}

void rtc_close(void){
	TIM_DeInit(TIM4);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM4, DISABLE);
	return;
}

#ifdef INSERT_TO_IRQ                    // stm32f4xx_it.c
#include <sys/time.h>
void TIM4_IRQHandler(void){
    extern volatile time_t    rtc_time;
    rtc_time++;
    return;
}
#endif

int settimeofday(const struct timeval *tv, const struct timezone *tz){
    int    r;

    if (tv != NULL) {
        rtc_time = tv->tv_sec;
        TIM_SetCounter(TIM4, tv->tv_usec / 10);
        r = 0;
    } else {
        r = -1;
        errno = EFAULT;
    }
    return(r);
}

int gettimeofday(struct timeval *tp, void *tzp){
    struct timezone    *tz;

    tz = tzp;
    if (tp != NULL) {
        tp->tv_sec  = rtc_time;
        tp->tv_usec = TIM_GetCounter(TIM4) * 10;
    }
    if (tz != NULL) {
        tz->tz_minuteswest = 0;
        tz->tz_dsttime     = 0;
    }
    return(0);
}
#endif

//
//  PIC32MX Implementation
//
#ifdef PIC32MX

//  Driver timer version
#ifdef RTC_DRV_TMR
#include "driver/tmr/drv_tmr.h"

uint32_t rtc_handler;
uint32_t rtc_period;

void rtc_callback(uintptr_t context, uint32_t alarmCount){
    rtc_time++;
    return;
}

void rtc_open(void) {
    rtc_time = 0;
    rtc_handler = DRV_TMR_Open(RTC_DRV_TMR, DRV_IO_INTENT_EXCLUSIVE);
    rtc_period  = DRV_TMR_CounterFrequencyGet(rtc_handler);
    DRV_TMR_AlarmRegister(rtc_handler, rtc_period, true, 0, rtc_callback);
    DRV_TMR_Start(rtc_handler);
    return;
}

void rtc_close(void){
    DRV_TMR_Stop(rtc_handler);
    DRV_TMR_Close(rtc_handler);
    return;
}

int settimeofday(const struct timeval *tv, void *tz){
    int    r;

    if (tv != NULL) {
        rtc_time = tv->tv_sec;
        DRV_TMR_CounterValueSet(rtc_handler, tv->tv_usec / 1.0e6 * rtc_period);
        r = 0;
    } else {
        r = -1;
        errno = EFAULT;
    }
    return(r);
}

int gettimeofday(struct timeval *tp, void *tzp){
    struct timezone    *tz;

    tz = tzp;
    if (tp != NULL) {
        tp->tv_sec  = rtc_time;
        tp->tv_usec = (float)DRV_TMR_CounterValueGet(rtc_handler) / (float)rtc_period * 1.0e6;
    }
    if (tz != NULL) {
        tz->tz_minuteswest = 0;
        tz->tz_dsttime     = 0;
    }
    return(0);
}

// Peripheral TMR version
#else

#include <stdint.h>
#include <sys/attribs.h>                // __ISR()
#include "system_config.h"              // "SYS_CLK_BUS_PERIPHERAL_1"
#include "peripheral/tmr/plib_tmr.h"    // Harmony peripheral librarys
#include "peripheral/int/plib_int.h"    // Harmony peripheral librarys

//  Config
#define MY_TMR_ID           TMR_ID_5
#define MY_TMR_SRC          INT_SOURCE_TIMER_5
#define MY_TMR_VEC          INT_VECTOR_T5
#define MY_TMR_ISR          _TIMER_5_VECTOR
#define MY_TMR_PRI          INT_PRIORITY_LEVEL6
#define MY_TMR_SUB          INT_SUBPRIORITY_LEVEL0
#define MY_TMR_IPL          IPL6AUTO
#define INTERVAL            1.0         // [s]
#define PERIOD              (SYS_CLK_BUS_PERIPHERAL_1 / (256/INTERVAL))

void __ISR(MY_TMR_ISR, MY_TMR_IPL) _IntHandlerRtcTmr(void) {
    rtc_time++;
    PLIB_INT_SourceFlagClear(INT_ID_0, MY_TMR_SRC);
    return;
}

void rtc_open(void) {
    rtc_time = 0;
    PLIB_TMR_ClockSourceSelect  (MY_TMR_ID, TMR_CLOCK_SOURCE_PERIPHERAL_CLOCK);
    PLIB_TMR_PrescaleSelect     (MY_TMR_ID, TMR_PRESCALE_VALUE_256);
    PLIB_TMR_Mode16BitEnable    (MY_TMR_ID);
    PLIB_TMR_Counter16BitClear  (MY_TMR_ID);
    PLIB_TMR_Period16BitSet     (MY_TMR_ID, PERIOD);
    PLIB_TMR_Start              (MY_TMR_ID);

    PLIB_INT_VectorPrioritySet   (INT_ID_0, MY_TMR_VEC, MY_TMR_PRI);
    PLIB_INT_VectorSubPrioritySet(INT_ID_0, MY_TMR_VEC, MY_TMR_SUB);
    PLIB_INT_SourceFlagClear     (INT_ID_0, MY_TMR_SRC);
    PLIB_INT_SourceEnable        (INT_ID_0, MY_TMR_SRC);
    return;
}

void rtc_close(void){
    PLIB_TMR_Stop(MY_TMR_ID);
    return;
}

int settimeofday(const struct timeval *tv, void *tz){
    int    r;

    if (tv != NULL) {
        rtc_time = tv->tv_sec;
        PLIB_TMR_Counter16BitSet(MY_TMR_ID, tv->tv_usec / 1.0e6 * PERIOD);
        r = 0;
    } else {
        r = -1;
        errno = EFAULT;
    }
    return(r);
}

int gettimeofday(struct timeval *tp, void *tzp){
    struct timezone    *tz;

    tz = tzp;
    if (tp != NULL) {
        tp->tv_sec  = rtc_time;
        tp->tv_usec = (float)PLIB_TMR_Counter16BitGet(MY_TMR_ID) / (float)PERIOD * 1.0e6;
    }
    if (tz != NULL) {
        tz->tz_minuteswest = 0;
        tz->tz_dsttime     = 0;
    }
    return(0);
}
#endif  // RTC_DRV_TMR

#endif  // PIC32MX
